"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.renderStack = renderStack;

var _chalk = _interopRequireDefault(require("chalk"));

var _fs = require("fs");

var _highlightEs = _interopRequireDefault(require("highlight-es"));

var _yamlifyObject = _interopRequireDefault(require("yamlify-object"));

var _prototype = require("./prototype");

var _stack = require("./stack");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/*
 *                            *** MIT LICENSE ***
 * -------------------------------------------------------------------------
 * This code may be modified and distributed under the MIT license.
 * See the LICENSE file for details.
 * -------------------------------------------------------------------------
 *
 * @summary
 *
 * @author    Alvis HT Tang <alvis@hilbert.space>
 * @license   MIT
 * @copyright Copyright (c) 2020 - All Rights Reserved.
 * -------------------------------------------------------------------------
 */

/** default number of lines from the targeted source line to be displayed */
const DEFAULT_SPREAD = 4;
const DEFAULT_CODE_THEME = {
  string: _chalk.default.green,
  punctuator: _chalk.default.grey,
  keyword: _chalk.default.cyan,
  number: _chalk.default.magenta,
  regex: _chalk.default.magenta,
  comment: _chalk.default.grey.bold,
  invalid: _chalk.default.inverse
};
const DEFAULT_YAML_THEME = {
  error: _chalk.default.red,
  symbol: _chalk.default.magenta,
  string: _chalk.default.green,
  date: _chalk.default.cyan,
  number: _chalk.default.magenta,
  boolean: _chalk.default.yellow,
  null: _chalk.default.yellow.bold,
  undefined: _chalk.default.yellow.bold
};
/**
 * render associations of an error
 * @param error the related error
 * @returns a rendered string to print
 */

function renderAssociations(error) {
  const blocks = [];

  if (error instanceof _prototype.Xception && error.namespace) {
    blocks.push(_chalk.default.blue.underline(error.namespace));
  }

  if (error instanceof _prototype.Xception && error.tags.length) {
    blocks.push(...error.tags.map(tag => _chalk.default.cyan.bold(tag)));
  }

  return blocks.length ? '\n    ' + blocks.join(' ') : null;
}
/**
 * render a description line
 * @param block a stack block about an error description
 * @param error error the related error
 * @returns a rendered string to print
 */


function renderDescription(block, error) {
  const {
    name,
    message
  } = block;

  const description = _chalk.default.red(`[${_chalk.default.bold(name)}] ${message}`);

  const association = renderAssociations(error);
  const meta = renderMeta(error);
  return [description, association, meta].filter(block => !!block).join('\n');
}
/**
 * render a location line
 * @param block a stack block about a location
 * @param options optional parameters
 * @param options.showSource indicate whether a source frame should be shown
 * @returns a rendered string to print
 */


function renderLocation(block, options) {
  const {
    entry,
    path,
    line,
    column
  } = block;
  const {
    showSource
  } = { ...options
  };
  const location = `${path}:${line}:${column}`;
  const sourceFrame = showSource ? renderSource(block) : '';
  return `    at ${_chalk.default.grey.bold(entry)} (${_chalk.default.grey.underline(location)})` + (sourceFrame ? '\n' + sourceFrame : '');
}
/**
 * render metadata in an error
 * @param error the related error
 * @returns a rendered string to print
 */


function renderMeta(error) {
  return error instanceof _prototype.Xception && Object.keys(error.meta).length ? (0, _yamlifyObject.default)(error.meta, {
    indent: '    ',
    prefix: '\n',
    postfix: '\n',
    colors: DEFAULT_YAML_THEME
  }) : null;
}
/**
 * render a source frame
 * @param block a location block
 * @param options options for rendering
 * @param options.spread the number of lines from the targeted source line to be displayed
 * @returns a rendered string to print
 */


function renderSource(block, options) {
  const {
    path,
    line
  } = block;
  const {
    spread = DEFAULT_SPREAD
  } = { ...options
  }; // no source frame if the source is missing

  if (!(0, _fs.existsSync)(path)) {
    return '';
  }

  const content = (0, _fs.readFileSync)(path).toString();
  const highlighted = (0, _highlightEs.default)(content, DEFAULT_CODE_THEME);
  const lines = highlighted.split(/[\r\n]/);
  const base = Math.max(line - spread - 1, 0);
  const displayLines = lines.slice(base, line + spread);
  const lineNumberWidth = (line + spread).toString().length;
  const sourceFrame = displayLines.map((source, index) => {
    const currentLine = base + index + 1;
    const formattedLine = currentLine.toString().padStart(lineNumberWidth);
    const isTarget = currentLine === line;
    const prefix = isTarget ? '>' : ' ';
    const gutter = ` ${prefix} ${formattedLine} `;
    return `${isTarget ? _chalk.default.bgRed(gutter) : gutter}| ${source}`;
  }).join('\n');
  return '\n' + sourceFrame + '\n';
}
/**
 * render a highly readable error stack
 * @param error the error to be rendered
 * @param options optional parameters
 * @param options.showSource indicate whether a source frame should be shown
 * @param options.filter a filter determining whether a stack should be shown given the file path
 * @returns a rendered string to print
 */


function renderStack(error, options) {
  const {
    showSource = false,
    filter = path => !path.includes('node:internal')
  } = { ...options
  };
  const blocks = (0, _stack.disassembleStack)(error.stack);
  let currentError = error;
  const renderedBlocks = [];

  for (let i = 0; i < blocks.length; i++) {
    const block = blocks[i];

    if (block.type === 'location' && filter(block.path)) {
      renderedBlocks.push(renderLocation(block, {
        showSource: // NOTE a location block must follow a description block
        showSource && blocks[i - 1].type === 'description'
      }));
    } else if (block.type === 'description') {
      renderedBlocks.push(renderDescription(block, currentError));
      currentError = error['cause'];
    }
  }

  return renderedBlocks.join('\n');
}
//# sourceMappingURL=render.js.map